__all__ = ["observable"]


class observable:

    def __init__(self):
        self.observers = []

    def add_observer(self, function):
        self.observers.append(function)

    def observer(self, function):
        self.add_observer(function)
        return function

    def notify(self, *args, **kwargs):
        for observer in self.observers:
            observer(*args, **kwargs)



import asyncio
import typing

__all__ = ["merge_streams", "make_closable"]


class close_stream(StopAsyncIteration):
    """
    Raise when you need to close an asynchronous stream.
    """


async def merge_streams(*streams):
    """
    Merges multiple asynchronous streams into a single stream.
    It takes any number of streams as arguments and continuously yields values from each stream.
    Close all streams when one of them is closed.
    """
    pending_tasks = [asyncio.create_task(anext(i)) for i in streams]

    def next_value(task):
        i = pending_tasks.index(task)
        s = streams[i]
        c = anext(s)
        pending_tasks[i] = asyncio.create_task(c)

    active = True
    while active:
        done_tasks, _ = await asyncio.wait(pending_tasks, return_when=asyncio.FIRST_COMPLETED)
        for task in done_tasks:
            result = task.result()
            if isinstance(result, close_stream):
                pending_tasks.remove(task)
                active = False
                continue
            yield result
            next_value(task)

    for task in pending_tasks:
        task.cancel()


def make_closable(
    stream,
    on_close: typing.Callable | None = None,
):
    """
    Makes an asynchronous stream closable.

    Args:
    - stream: is an asynchronous stream that you want to make closable.
    - on_close: is a callable that takes no arguments and returns an awaitable that completes when the stream is closed.
    """
    close_event = asyncio.Event()

    async def on_close_stream():
        await close_event.wait()
        if callable(on_close):
            result = on_close()
            if asyncio.iscoroutine(result):
                await result
        yield close_stream()

    new_stream = merge_streams(stream, on_close_stream())
    return new_stream, close_event.set



import logging
import re
import typing

from . import _shared

if typing.TYPE_CHECKING:
    from . import _session
    from . import _service

__all__ = [
    "transition",
    "observable_state",
    "new_observable_state",
    "next_observer",
]


_logger = logging.getLogger(__name__)
_logger.setLevel(logging.INFO)


@_shared.dataclass
class transition:
    label: str
    sources: set["observable_state"]
    target: "observable_state"
    procedure: typing.Callable
    description: str | None = None
    priority: int = -1
    is_observer: bool = False

    def __post_init__(self):
        self.__name__ = self.label
        self.__doc__ = self.description
        self.is_observer = self.priority > -1

    async def __call__(
        self,
        payload: typing.Any = None,
        *,
        session: "_session.Almanet",
    ):
        result = await self.procedure(payload, session=session, transition=self)
        session.call(self.target._next_procedure.uri, result)
        return result


_state_label_re = re.compile("[A-Za-z0-9_]+")


@_shared.dataclass(slots=True)
class _state:
    service: "_service.service"
    label: str
    description: str | None = None
    _transitions: typing.List[transition] = ...

    def __post_init__(self):
        if not (isinstance(self.label, str) and len(self.label) > 0 and _state_label_re.match(self.label)):
            raise ValueError("`label` must contain uppercase words separated by underscore")
        self._transitions = []

    def _add_transition(
        self,
        sources: typing.Iterable["observable_state"],
        procedure: typing.Callable,
        label: str | None = None,
        description: str | None = None,
        **extra,
    ):
        if not callable(procedure):
            raise ValueError("decorated function must be callable")

        if label is None:
            label = procedure.__name__

        if description is None:
            description = procedure.__doc__

        if not all(isinstance(i, observable_state) for i in sources):
            raise ValueError(f"{label}: `target` must be `observable_state` instance")

        instance = transition(
            label=label,
            description=description,
            sources=set(sources),
            target=self,  # type: ignore
            procedure=procedure,
            **extra,
        )

        for i in sources:
            i._transitions.append(instance)
            i._transitions = sorted(i._transitions, key=lambda o: -o.priority)

        return instance

    def transition_from(
        self,
        *sources: "observable_state",
        **extra,
    ):
        def wrap(function):
            return self._add_transition(sources, procedure=function, **extra)

        return wrap


class priorities:
    unset = -1
    lowest = 0
    low = 25
    medium = 50
    high = 75
    highest = 100


class flow_execution_error(Exception): ...


@_shared.dataclass(slots=True)
class observable_state(_state):
    _next_procedure: "_service.procedure_model" = ...

    @property
    def observers(self) -> list[transition]:
        return [i for i in self._transitions if i.is_observer]

    async def next(
        self,
        payload: typing.Any = None,
        *,
        session: "_session.Almanet",
    ) -> typing.Any:
        _logger.debug(f"{self.label} begin")

        for observer in self.observers:
            _logger.debug(f"trying to call {observer.label} observer")
            try:
                result = await observer(payload=payload, session=session)
                _logger.debug(f"{observer.label} observer end")
                return result
            except Exception as e:
                _logger.error(f"during execution of {observer.label}: {repr(e)}")

        if len(self.observers) > 0:
            raise flow_execution_error(self.label)

    def __post_init__(self):
        super(observable_state, self).__post_init__()
        self._next_procedure = self.service.add_procedure(
            self.next,
            path=self.label,
            include_to_api=False,
            validate=False,
        )

    def _add_observer(
        self,
        priority: int = priorities.medium,
        **kwargs,
    ):
        if not (isinstance(priority, int) and priority > -1):
            raise ValueError("`priority` must be `integer` and greater than -1")

        existing_priorities = {i.priority for i in self.observers}
        if priority in existing_priorities:
            raise ValueError(f"invalid observer `{self.label}` (priority {priority} already exists)")

        return self._add_transition(**kwargs, priority=priority)

    def observe(
        self,
        *sources: "observable_state",
        **extra,
    ):
        def wrap(function):
            return self._add_observer(
                sources=sources,
                procedure=function,
                **extra,
            )

        return wrap

    def __hash__(self) -> int:
        return hash(self.label)


new_observable_state = observable_state


class next_observer(Exception):
    def __init__(self, reason: str, *args):
        if not isinstance(reason, str):
            raise ValueError("`reason` must be `string`")
        self.reason = reason
        self.args = args



from dataclasses import dataclass, field

from ._concurrent_context import *
from ._decoding import *
from ._encoding import *
from ._new_id import *
from ._observable import *
from ._schema import *
from ._streaming import *
from ._background_tasks import *
from ._validate_execution import *
from ._is_valid_uri import *

__all__ = [
    "dataclass",
    "field",
    *_concurrent_context.__all__,
    *_decoding.__all__,
    *_encoding.__all__,
    *_new_id.__all__,
    *_observable.__all__,
    *_schema.__all__,
    *_streaming.__all__,
    *_background_tasks.__all__,
    *_validate_execution.__all__,
    *_is_valid_uri.__all__,
]



import asyncio

from . import _session
from . import _shared

__all__ = [
    "session_pool",
    "new_session_pool",
    "acquire_active_session",
]


class session_pool:

    def __init__(self):
        self.joined = False
        self.sessions: list[_session.Almanet] = []

    async def spawn(
        self,
        addresses: tuple[str, ...],
        number_of_sessions: int = 1,
    ) -> None:
        if number_of_sessions < 1:
            raise ValueError("number_of_sessions must be greater than 0")

        for _ in range(number_of_sessions):
            session = _session.new_session()
            await session.join(*addresses)
            self.sessions.append(session)

        self.joined = True

        _current_session_pool.set(self)

    async def kill(
        self,
        *args,
        **kwargs,
    ) -> None:
        sessions, self.sessions = self.sessions, []
        async with asyncio.TaskGroup() as g:
            for i in sessions:
                coroutine = i.leave(*args, **kwargs)
                g.create_task(coroutine)

        self.joined = False

        _current_session_pool.set(None)

    def rotate(self) -> _session.Almanet:
        session = self.sessions.pop(0)
        self.sessions.append(session)
        return session


new_session_pool = session_pool


_current_session_pool = _shared.new_concurrent_context()


def acquire_active_session() -> _session.Almanet:
    pool = _current_session_pool.get(None)
    if pool is None:
        return _session.get_active_session()
    return pool.rotate()



import typing

import pydantic
import pydantic_core

__all__ = [
    "serialize",
    "serialize_json",
]


def serialize[T: typing.Any](
    annotation: type[T] | typing.Any = ...,
) -> typing.Callable[[typing.Any], T]:
    if annotation is ...:
        return lambda v: v

    model = pydantic.TypeAdapter(annotation)
    return lambda v: model.validate_python(v)


def serialize_json[T: typing.Any](
    annotation: type[T] | typing.Any = ...,
) -> typing.Callable[[bytes | str], T]:
    if annotation is ...:
        return lambda v: pydantic_core.from_json(v)

    model = pydantic.TypeAdapter(annotation)
    return lambda v: model.validate_json(v)



import asyncio
import signal

from . import _service
from . import _session_pool

__all__ = ["serve"]


def serve(
    *addresses: str,
    services: list[_service.service] = [],
) -> None:
    for service in services:
        if not isinstance(service, _service.service):
            raise ValueError("must be an instance of service")

    session_pool = _session_pool.new_session_pool()

    async def begin() -> None:
        await session_pool.spawn(addresses, len(services))
        for s in services:
            s._post_join_event.notify(session_pool)

    async def end() -> None:
        await session_pool.kill()
        loop.stop()

    loop = asyncio.get_event_loop()

    for s in (signal.SIGINT, signal.SIGTERM):
        loop.add_signal_handler(s, lambda: loop.create_task(end()))

    loop.create_task(begin())
    if not loop.is_running():
        loop.run_forever()



import asyncio
import logging
import typing

import pydantic_core

from . import _shared

if typing.TYPE_CHECKING:
    from . import _service

__all__ = [
    "client_iface",
    "invoke_event_model",
    "qmessage_model",
    "reply_event_model",
    "rpc_error",
    "Almanet",
    "new_session",
]


logger = logging.getLogger("almanet")
logger.setLevel(logging.INFO)


@_shared.dataclass(slots=True)
class qmessage_model[T: typing.Any]:
    """
    Represents a message in the queue.
    """

    id: str
    timestamp: int
    body: T
    attempts: int
    commit: typing.Callable[[], typing.Awaitable[None]]
    rollback: typing.Callable[[], typing.Awaitable[None]]


type returns_consumer[T: typing.Any] = tuple[typing.AsyncIterable[qmessage_model[T]], typing.Callable[[], None]]


class client_iface(typing.Protocol):
    """
    Interface for a client library.
    """

    async def connect(
        self,
        addresses: typing.Sequence[str],
    ) -> None:
        raise NotImplementedError()

    async def produce(
        self,
        topic: str,
        message: str | bytes,
    ) -> None:
        raise NotImplementedError()

    async def consume(
        self,
        topic: str,
        channel: str,
    ) -> returns_consumer[bytes]:
        raise NotImplementedError()

    async def close(self) -> None:
        raise NotImplementedError()


@_shared.dataclass(slots=True)
class invoke_event_model:
    """
    Represents an invocation event.
    """

    id: str
    caller_id: str
    payload: typing.Any
    reply_topic: str

    @property
    def expired(self) -> bool:
        # TODO
        return False


@_shared.dataclass(slots=True)
class reply_event_model:
    """
    Represents a reply event.
    """

    call_id: str
    is_error: bool
    payload: typing.Any


class rpc_error(Exception):
    """
    Represents an RPC error.
    You can inherit from this class to create your own error.
    """

    __slots__ = ("name", "args")

    def __init__(
        self,
        *args,
        name: str | None = None,
    ) -> None:
        self.name = name or self.__class__.__name__
        self.args = args

    def __str__(self) -> str:
        return f"{self.name}{self.args}"


@_shared.dataclass(slots=True)
class registration_model:
    """
    Represents a registered procedure to call.
    """

    uri: str
    channel: str
    procedure: typing.Callable
    session: "Almanet"

    @property
    def __name__(self):
        return self.uri

    async def execute(
        self,
        invocation: invoke_event_model,
    ) -> reply_event_model:
        __log_extra = {"registration": str(self), "invocation": str(invocation)}
        try:
            logger.debug("trying to execute procedure", extra=__log_extra)
            reply_payload = await self.procedure(
                invocation.payload,
                session=self.session,
            )
            return reply_event_model(call_id=invocation.id, is_error=False, payload=reply_payload)
        except Exception as e:
            if isinstance(e, rpc_error):
                error_name = e.name
                error_message = e.args
            elif isinstance(e, pydantic_core.ValidationError):
                error_name = "ValidationError"
                error_message = repr(e)
            else:
                error_name = "InternalError"
                error_message = "oops"
                logger.exception("during execute procedure", extra=__log_extra)
            return reply_event_model(
                call_id=invocation.id,
                is_error=True,
                payload={"name": error_name, "message": error_message},
            )


class Almanet:
    """
    Represents a session, connected to message broker.
    """

    def __init__(
        self,
        *addresses: str,
        client_class: type[client_iface] | None = None,
    ) -> None:
        self.id = _shared.new_id()
        self.addresses: tuple[str, ...] = addresses
        self.joined = False
        if client_class is None:
            from . import _clients

            client_class = _clients.DEFAULT_CLIENT
        self._client: client_iface = client_class()
        self._background_tasks = _shared.background_tasks()
        self._post_join_event = _shared.observable()
        self._leave_event = _shared.observable()
        self._pending_replies: typing.MutableMapping[str, asyncio.Future[reply_event_model]] = {}
        self._invocations_relay = asyncio.Event()
        self._invocations_relay.set()

    @property
    def version(self) -> float:
        return 0.1

    async def _produce(
        self,
        uri: str,
        payload: typing.Any,
    ) -> None:
        try:
            message_body = _shared.dump(payload)
        except Exception as e:
            logger.error(f"during encode payload: {repr(e)}")
            raise e

        try:
            logger.debug(f"trying to produce {uri} topic")
            await self._client.produce(uri, message_body)
        except Exception as e:
            logger.exception(f"during produce {uri} topic")
            raise e

    def produce(
        self,
        uri: str,
        payload: typing.Any,
    ) -> asyncio.Task[None]:
        """
        Produce a message with a specified topic and payload.
        """
        return self._background_tasks.schedule(self._produce(uri, payload))

    async def _serialize[T: typing.Any](
        self,
        messages_stream: typing.AsyncIterable[qmessage_model[bytes]],
        payload_model: type[T] | typing.Any = ...,
    ) -> typing.AsyncIterable[qmessage_model[T]]:
        serializer = _shared.serialize_json(payload_model)

        async for message in messages_stream:
            try:
                message.body = serializer(message.body)
            except:
                logger.exception("during decode payload")
                continue

            yield message  # type: ignore

    async def consume[T](
        self,
        topic: str,
        channel: str,
        *,
        payload_model: type[T] | typing.Any = ...,
    ) -> returns_consumer[T]:
        """
        Consume messages from a message broker with the specified topic and channel.
        It returns a tuple of a stream of messages and a function that can stop consumer.
        """
        logger.debug(f"trying to consume {topic}/{channel}")

        messages_stream, stop_consumer = await self._client.consume(topic, channel)
        self._leave_event.add_observer(stop_consumer)

        messages_stream = self._serialize(messages_stream, payload_model)

        return messages_stream, stop_consumer

    async def _consume_replies(
        self,
        ready_event: asyncio.Event,
    ) -> None:
        messages_stream, _ = await self.consume(
            f"_rpc_._reply_.{self.id}",
            channel="rpc-recipient",
        )
        logger.debug("reply event consumer begin")
        ready_event.set()
        async for message in messages_stream:
            __log_extra = {"incoming_message": str(message)}
            try:
                reply = reply_event_model(**message.body)
                __log_extra["reply"] = str(reply)
                logger.debug("new reply", extra=__log_extra)

                pending = self._pending_replies.get(reply.call_id)
                if pending is None:
                    logger.warning("pending event not found", extra=__log_extra)
                else:
                    pending.set_result(reply)
            except:
                logger.exception("during parse reply", extra=__log_extra)

            await message.commit()
            logger.debug("successful commit", extra=__log_extra)
        logger.debug("reply event consumer end")

    _call_only_args = tuple[str, typing.Any]

    class _call_only_kwargs(typing.TypedDict):
        timeout: typing.NotRequired[int]

    async def _call_only(
        self,
        *args: typing.Unpack[_call_only_args],
        **kwargs: typing.Unpack[_call_only_kwargs],
    ) -> reply_event_model:
        uri, payload = args
        timeout = kwargs.get("timeout") or 60

        invocation = invoke_event_model(
            id=_shared.new_id(),
            caller_id=self.id,
            payload=payload,
            reply_topic=f"_rpc_._reply_.{self.id}",
        )

        __log_extra = {"uri": uri, "timeout": timeout, "invoke_event": str(invocation)}
        logger.debug("trying to call", extra=__log_extra)

        pending_reply_event = asyncio.Future[reply_event_model]()
        self._pending_replies[invocation.id] = pending_reply_event

        try:
            async with asyncio.timeout(timeout):
                await self.produce(f"_rpc_.{uri}", invocation)

                reply_event = await pending_reply_event
                __log_extra["reply_event"] = str(reply_event)
                logger.debug("new reply event", extra=__log_extra)

                if reply_event.is_error:
                    raise rpc_error(
                        reply_event.payload["message"],
                        name=reply_event.payload["name"],
                    )

                return reply_event
        except Exception as e:
            logger.error(f"during call {uri}", extra={**__log_extra, "error": repr(e)})
            raise e
        finally:
            self._pending_replies.pop(invocation.id)

    def call_only(
        self,
        *args: typing.Unpack[_call_only_args],
        **kwargs: typing.Unpack[_call_only_kwargs],
    ) -> asyncio.Task[reply_event_model]:
        """
        Executes the remote procedure using the payload.
        """
        return self._background_tasks.schedule(self._call_only(*args, **kwargs))

    class _call_kwargs[R](_call_only_kwargs):
        result_model: typing.NotRequired[type[R]]

    async def _call[I, O](
        self,
        topic: typing.Union[str, "_service.procedure_model[I, O]"],
        payload: I,
        **kwargs: typing.Unpack[_call_kwargs],
    ) -> O:
        result_model = kwargs.pop("result_model", ...)

        if isinstance(topic, str):
            uri = topic
        else:
            uri = topic.uri
            if result_model is ...:
                result_model = topic.return_model

        serialize_result = _shared.serialize(result_model)

        reply_event = await self._call_only(uri, payload, **kwargs)

        try:
            result = serialize_result(reply_event.payload)
        except pydantic_core.ValidationError as e:
            logger.error(f"invalid result from {uri}", extra={"error": repr(e)})
            raise e

        return result

    @typing.overload
    def call[I, O](
        self,
        topic: "_service.procedure_model[I, O]",
        payload: I,
        **kwargs: typing.Unpack[_call_kwargs],
    ) -> asyncio.Task[O]: ...

    @typing.overload
    def call[O](
        self,
        topic: str,
        payload: typing.Any,
        **kwargs: typing.Unpack[_call_kwargs[O]],
    ) -> asyncio.Task[O]: ...

    def call(self, *args, **kwargs):
        """
        Executes the remote procedure using the payload.
        Returns a instance of result model.
        """
        return self._background_tasks.schedule(self._call(*args, **kwargs))

    async def _multicall_only(
        self,
        *args: typing.Unpack[_call_only_args],
        **kwargs: typing.Unpack[_call_only_kwargs],
    ) -> list[reply_event_model]:
        uri, payload = args
        timeout = kwargs.get("timeout") or 60

        invocation = invoke_event_model(
            id=_shared.new_id(),
            caller_id=self.id,
            payload=payload,
            reply_topic=f"_rpc_._replies_.{self.id}",
        )

        __log_extra = {"uri": uri, "timeout": timeout, "invoke_event": str(invocation)}

        messages_stream, stop_consumer = await self.consume(invocation.reply_topic, "rpc-recipient")

        result = []
        try:
            async with asyncio.timeout(timeout):
                await self.produce(f"_rpc_.{uri}", invocation)

                async for message in messages_stream:
                    try:
                        logger.debug("new reply event", extra=__log_extra)
                        reply = reply_event_model(**message.body)
                        result.append(reply)
                    except:
                        logger.exception("during parse reply event", extra=__log_extra)

                    await message.commit()
        except TimeoutError:
            stop_consumer()

        logger.debug(f"multicall {uri} done")

        return result

    def multicall_only(
        self,
        *args: typing.Unpack[_call_only_args],
        **kwargs: typing.Unpack[_call_only_kwargs],
    ) -> asyncio.Task[list[reply_event_model]]:
        """
        Execute simultaneously multiple procedures using the payload.
        """
        return self._background_tasks.schedule(self._multicall_only(*args, **kwargs))

    async def _handle_invocation(
        self,
        registration: registration_model,
        message: qmessage_model,
    ):
        __log_extra = {"registration": str(registration), "incoming_message": str(message)}
        try:
            invocation = invoke_event_model(**message.body)
            __log_extra["invocation"] = str(invocation)
            logger.debug("new invocation", extra=__log_extra)

            if invocation.expired:
                logger.warning("invocation expired", extra=__log_extra)
            else:
                reply = await registration.execute(invocation)
                logger.debug("trying to reply", extra=__log_extra)
                await self.produce(invocation.reply_topic, reply)
        except:
            logger.exception("during execute invocation", extra=__log_extra)
        finally:
            await message.commit()
            logger.debug("successful commit", extra=__log_extra)

    async def _consume_invocations(
        self,
        registration: registration_model,
    ) -> None:
        logger.debug(f"trying to register {registration.uri}/{registration.channel}")
        messages_stream, _ = await self.consume(f"_rpc_.{registration.uri}", registration.channel)
        async for message in messages_stream:
            self._background_tasks.schedule(self._handle_invocation(registration, message))
            await self._invocations_relay.wait()
        logger.debug(f"consumer {registration.uri} down")

    def register(
        self,
        topic: str,
        procedure: typing.Callable,
        *,
        channel: str = "main",
    ) -> registration_model:
        """
        Register a procedure with a specified topic and payload.
        Returns the created registration.
        """
        if not self.joined:
            raise RuntimeError(f"session {self.id} not joined")

        registration = registration_model(
            uri=topic,
            channel=channel,
            procedure=procedure,
            session=self,
        )

        self._background_tasks.schedule(self._consume_invocations(registration), daemon=True)

        return registration

    async def join(
        self,
        *addresses: str,
    ) -> None:
        """
        Join the session to message broker.
        """
        if self.joined:
            raise RuntimeError(f"session {self.id} already joined")

        self.addresses += addresses

        if len(self.addresses) == 0:
            raise ValueError("at least one address must be specified")

        if not all(isinstance(i, str) for i in self.addresses):
            raise ValueError("addresses must be a iterable of strings")

        logger.debug(f"trying to connect addresses={self.addresses}")

        await self._client.connect(self.addresses)

        consume_replies_ready = asyncio.Event()
        self._background_tasks.schedule(
            self._consume_replies(consume_replies_ready),
            daemon=True,
        )
        await consume_replies_ready.wait()

        self.joined = True
        self._post_join_event.notify()
        logger.info(f"session {self.id} joined")

    async def __aenter__(self) -> "Almanet":
        if not self.joined:
            await self.join(*self.addresses)

        _active_session.set(self)

        return self

    async def leave(
        self,
        reason: str | None = None,
    ) -> None:
        """
        Leave the session from message broker.
        """
        if not self.joined:
            raise RuntimeError(f"session {self.id} not joined")
        self.joined = False

        logger.debug(f"trying to leave {self.id} session, reason: {reason}")

        self._invocations_relay.clear()

        logger.debug(f"session {self.id} await task pool complete")
        await self._background_tasks.complete()

        self._leave_event.notify()

        self._invocations_relay.set()

        logger.debug(f"session {self.id} trying to close connection")
        await self._client.close()

        # some tasks have not been completed yet
        await asyncio.sleep(1)

        logger.warning(f"session {self.id} left")

    async def __aexit__(self, etype, evalue, etraceback) -> None:
        _active_session.set(None)

        if self.joined:
            await self.leave()


new_session = Almanet


_active_session = _shared.new_concurrent_context()


def get_active_session() -> Almanet:
    session = _active_session.get(None)
    if session is None:
        raise RuntimeError("active session not found")
    return session



import typing

import pydantic


__all__ = [
    "extract_annotations",
    "generate_json_schema",
    "describe_function",
]


def extract_annotations(
    function: typing.Callable,
    payload_annotation=...,
    return_annotation=...,
):
    if payload_annotation is ...:
        payload_annotation = function.__annotations__.get("payload", ...)
    if return_annotation is ...:
        return_annotation = function.__annotations__.get("return", ...)
    return payload_annotation, return_annotation


def generate_json_schema(annotation):
    """
    Generates a JSON schema from an annotation.
    """
    if annotation is ...:
        return None

    model = pydantic.TypeAdapter(annotation)
    return model.json_schema()


def describe_function(
    target: typing.Callable,
    description: str | None = None,
    payload_annotation=...,
    return_annotation=...,
):
    """
    Generates a JSON schema from a function.
    """
    if description is None:
        description = target.__doc__
    payload_annotation, return_annotation = extract_annotations(target, payload_annotation, return_annotation)
    payload_json_schema = generate_json_schema(payload_annotation)
    return_json_schema = generate_json_schema(return_annotation)
    return {
        "description": description,
        "payload_json_schema": payload_json_schema,
        "return_json_schema": return_json_schema,
    }



import typing

from . import _shared
from . import _session_pool

__all__ = [
    "service",
    "new_service",
]


class _function[I, O](typing.Protocol):
    __name__: str

    async def __call__(
        self,
        payload: I,
        *args,
        **kwargs,
    ) -> O: ...


@_shared.dataclass
class procedure_model[I, O]:
    service: "service"
    function: _function[I, O]
    path: str = ...
    description: str | None = None
    tags: set[str] = ...
    include_to_api: bool = True
    validate: bool = True
    payload_model: typing.Any = ...
    return_model: typing.Any = ...
    _has_implementation: bool = False
    _schema: typing.Mapping = ...

    def __post_init__(self):
        if not callable(self.function):
            raise ValueError("decorated function must be callable")
        if not isinstance(self.path, str):
            self.path = self.function.__name__
        self.payload_model, self.return_model = _shared.extract_annotations(
            self.function, self.payload_model, self.return_model
        )
        self._schema = _shared.describe_function(
            self.function,
            self.description,
            payload_annotation=self.payload_model,
            return_annotation=self.return_model,
        )
        self.description = self._schema["description"]
        if self.validate:
            self.function = _shared.validate_execution(self.function, self.payload_model, self.return_model)
        if self.tags is ...:
            self.tags = set()

    @property
    def uri(self):
        return '.'.join([self.service.pre, self.path])

    def __call__(self, payload: I) -> typing.Awaitable[O]:
        session = _session_pool.acquire_active_session()

        if self._has_implementation:
            return self.function(payload, session=session)

        return session.call(self.uri, payload, result_model=self.return_model)

    def implements(
        self,
        real_function: _function[I, O],
    ) -> "procedure_model[I, O]":
        if self._has_implementation:
            raise ValueError("procedure already implemented")
        self._has_implementation = True

        procedure = self.service.add_procedure(
            real_function,
            path=self.path,
            include_to_api=self.include_to_api,
            description=self.description,
            tags=self.tags,
            validate=self.validate,
            payload_model=self.payload_model,
            return_model=self.return_model,
        )
        return procedure


class service:
    def __init__(
        self,
        prepath: str = '',
        tags: set[str] | None = None,
    ) -> None:
        self.channel = "service"
        self.pre: str = prepath
        self.default_tags: set[str] = set(tags or [])
        self.procedures: list[procedure_model] = []
        self.background_tasks = _shared.background_tasks()
        self._post_join_event = _shared.observable()
        self._post_join_event.add_observer(self._share_all)

    @property
    def routes(self) -> set[str]:
        return {f"{i.uri}:{self.channel}" for i in self.procedures}

    def post_join[T: typing.Callable](
        self,
        function: T,
    ) -> T:
        def decorator(
            session_pool: "_session_pool.session_pool",
            *args,
            **kwargs,
        ):
            session = session_pool.rotate()
            coroutine = function(session, *args, **kwargs)
            self.background_tasks.schedule(coroutine)

        self._post_join_event.add_observer(decorator)
        return function

    class _register_procedure_kwargs(typing.TypedDict):
        path: typing.NotRequired[str]
        include_to_api: typing.NotRequired[bool]
        description: typing.NotRequired[str | None]
        tags: typing.NotRequired[set[str]]
        validate: typing.NotRequired[bool]
        payload_model: typing.NotRequired[typing.Any]
        return_model: typing.NotRequired[typing.Any]

    @typing.overload
    def public_procedure[I, O](
        self,
        function: _function[I, O],
    ) -> procedure_model[I, O]: ...

    @typing.overload
    def public_procedure[I, O](
        self,
        **kwargs: typing.Unpack[_register_procedure_kwargs],
    ) -> typing.Callable[[_function[I, O]], procedure_model[I, O]]: ...

    def public_procedure(
        self,
        function = None,
        **kwargs: typing.Unpack[_register_procedure_kwargs],
    ) -> procedure_model | typing.Callable[[_function], procedure_model]:
        if function is None:
            return lambda function: procedure_model(self, function, **kwargs)
        return procedure_model(self, function, **kwargs)

    def add_procedure(
        self,
        function: typing.Callable,
        **kwargs: typing.Unpack[_register_procedure_kwargs],
    ) -> procedure_model:
        procedure = procedure_model(
            self,
            function,
            **kwargs,
            _has_implementation=True,
        )
        self.procedures.append(procedure)
        return procedure

    @typing.overload
    def procedure[I, O](
        self,
        **kwargs: typing.Unpack[_register_procedure_kwargs],
    ) -> typing.Callable[[_function[I, O]], procedure_model[I, O]]: ...

    @typing.overload
    def procedure[I, O](
        self,
        function: _function[I, O],
    ) -> procedure_model[I, O]: ...

    def procedure(
        self,
        function = None,
        **kwargs: typing.Unpack[_register_procedure_kwargs],
    ) -> procedure_model | typing.Callable[[_function], procedure_model]:
        """
        Allows you to easily add procedures (functions) to a microservice by using a decorator.
        Returns a decorated function.
        """
        if function is None:
            return lambda function: self.add_procedure(function, **kwargs)
        return self.add_procedure(function, **kwargs)

    def _share_self_schema(
        self,
        **extra,
    ) -> None:
        async def procedure(*args, **kwargs):
            return {
                "session_id": self.session.id,
                "session_version": self.session.version,
                "routes": list(self.routes),
                **extra,
            }

        self.session.register(
            "_schema_.client",
            procedure,
            channel=self.session.id,
        )

    def _share_procedure_schema(
        self,
        registration: procedure_model,
    ) -> None:
        tags = registration.tags | self.default_tags
        if len(tags) == 0:
            tags = {"Default"}

        async def procedure(*args, **kwargs):
            return {
                "session_id": self.session.id,
                "session_version": self.session.version,
                "uri": registration.uri,
                "validate": registration.validate,
                "payload_model": registration.payload_model,
                "return_model": registration.return_model,
                "tags": tags,
                **registration._schema,
            }

        self.session.register(
            f"_schema_.{registration.uri}.{self.channel}",
            procedure,
            channel=self.channel,
        )

    def _share_all(
        self,
        session_pool: "_session_pool.session_pool",
    ) -> None:
        self.session = session_pool.rotate()

        self._share_self_schema()

        for procedure in self.procedures:
            for session in session_pool.sessions:
                session.register(
                    procedure.uri,
                    procedure.function,
                    channel=self.channel,
                )

            if procedure.include_to_api:
                self._share_procedure_schema(procedure)


new_service = service



from . import _clients as clients
from . import _shared as shared

from ._session import *
from ._flow import *
from ._module import *
from ._service import *

__all__ = [
    "clients",
    "shared",
    *_flow.__all__,
    *_module.__all__,
    *_service.__all__,
    *_session.__all__,
]



import re

__all__ = ["is_valid_uri"]

_uri_re = re.compile("^[a-zA-Z_][a-zA-Z0-9_]*$")


def is_valid_uri(v: str) -> bool:
    """
    Validate a URI string.

    The URI segment must:
    - have a length greater than 0
    - only contain Latin symbols and numbers
    - can include '.' and '_' symbols
    - not allow numbers as the first symbol

    Valid URIs:
    - "net.example.v2.greet"
    - "net.example.v2._test_greet"

    Args:
    - v: The URI string to validate
    """
    if not isinstance(v, str):
        return False

    segments = v.split('.')
    for s in segments:
        if not _uri_re.match(s):
            return False

    return True



import asyncio

from . import _decoding
from . import _schema

__all__ = ["validate_execution"]


def validate_execution(
    function,
    payload_model=...,
    return_model=...,
):
    """
    Takes a function as input and returns a decorator.
    The decorator validates the input payload and output return of the function based on their annotations.

    Args:
    - function: the function to decorate with validator
    - payload_model: the model of the input
    - return_model: the model of the output
    """
    payload_model, return_model = _schema.extract_annotations(function, payload_model, return_model)
    payload_validator = _decoding.serialize(payload_model)
    return_validator = _decoding.serialize(return_model)

    async def decorator(payload, *args, **kwargs):
        # FIXME reraise custom validation error
        payload = payload_validator(payload)
        result = function(payload, *args, **kwargs)
        if asyncio.iscoroutine(result):
            result = await result
        # FIXME reraise custom validation error
        return return_validator(result)

    return decorator



import contextvars

from . import _new_id

__all__ = ["new_concurrent_context"]


def new_concurrent_context[T](*args, **kwargs) -> contextvars.ContextVar[T]:
    name = _new_id.new_id()
    return contextvars.ContextVar(name, *args, **kwargs)



import typing

import pydantic

__all__ = [
    "dump",
]


def dump(v: typing.Any) -> bytes:
    if isinstance(v, bytes):
        return v

    if isinstance(v, str):
        return v.encode()

    codec = pydantic.RootModel(v)
    json_string = codec.model_dump_json()
    return json_string.encode()



import typing

import ansq

from almanet import _session
from almanet import _shared

if typing.TYPE_CHECKING:
    from ansq.tcp.types import NSQMessage

__all__ = ["ansq_client"]


class ansq_client:

    async def connect(
        self,
        addresses: typing.Sequence[str],
    ) -> None:
        self.addresses = addresses
        self.writer = await ansq.create_writer(
            nsqd_tcp_addresses=addresses,
        )

    async def close(self) -> None:
        await self.writer.close()

    async def produce(
        self,
        topic: str,
        message: str | bytes,
    ) -> None:
        await self.writer.pub(topic, message)

    async def _convert_ansq_message(
        self,
        ansq_messages_stream: typing.AsyncIterable['NSQMessage'],
    ) -> typing.AsyncIterable[_session.qmessage_model[bytes]]:
        async for ansq_message in ansq_messages_stream:
            almanet_message = _session.qmessage_model(
                id=ansq_message.id,
                timestamp=ansq_message.timestamp,
                body=ansq_message.body,
                attempts=ansq_message.attempts,
                commit=ansq_message.fin,
                rollback=ansq_message.req,
            )
            yield almanet_message

    async def consume(
        self,
        topic: str,
        channel: str,
    ) -> _session.returns_consumer[bytes]:
        reader = await ansq.create_reader(
            nsqd_tcp_addresses=self.addresses,
            topic=topic,
            channel=channel,
            connection_options=ansq.ConnectionOptions()
        )
        messages_stream = reader.messages()
        messages_stream = self._convert_ansq_message(messages_stream)
        # ansq does not close stream automatically
        return _shared.make_closable(messages_stream, reader.close)



import asyncio
import typing

__all__ = ["background_tasks"]


class background_tasks:
    """
    `asyncio.create_task` documentation has important note
    about avoiding a task disappearing mid-execution
    https://docs.python.org/3/library/asyncio-task.html#asyncio.create_task
    https://github.com/python/cpython/issues/88831
    """

    def __init__(self):
        self._tasks = set[asyncio.Task]()
        self._daemons = set[asyncio.Task]()

    def schedule(
        self,
        coroutine: typing.Coroutine,
        *,
        daemon: bool = False,
    ) -> asyncio.Task:
        """
        Schedules a coroutine for execution in the event loop.

        Args:
        - coroutine: the coroutine to be scheduled as an asyncio task.
        - daemon: marks the task as a daemon and does not wait completion.
        """
        task = asyncio.create_task(coroutine)
        task.add_done_callback(lambda _: self._tasks.discard(task))
        self._tasks.add(task)
        if daemon:
            self._daemons.add(task)
        return task

    async def complete(self) -> None:
        """
        Awaits the completion of all non-daemon tasks in the pool.
        This method will block until all non-daemon tasks have finished executing.
        """
        pending_tasks = self._tasks - self._daemons
        if len(pending_tasks) > 0:
            await asyncio.wait(pending_tasks)
